/* This file contains auxiliary predicates for device */
/* ECLiPSe Prolog version */



% This predicate appends multiple lists, which are given as the first
% argument, which is a list of lists!
append_lists([],[]) :- !.
append_lists([HL|TL],AL) :- 
	append_lists(TL,ATL),
	append(HL,ATL,AL).

% This predicate tests if the first list is a subset of the second.
% The order is not relevant (in contrast to ECLiPSe's subset/2)
my_subset([],_) :- !.
my_subset([H1|T1],L2) :- 
	member(H1,L2),
	my_subset(T1,L2).

% This predicate tests whether either of sets S1 or S2 is a subset of each other.
either_way_subset(S1,S2) :- 
	my_subset(S1,S2), !.
either_way_subset(S1,S2) :- 
	my_subset(S2,S1).

% This predicate subtracts the second list from the first, but the
% occurences of the second list are removed only once from the first
% list.
my_subtract(List1,[],List1) :- !.
my_subtract(List1,[H|T],List2) :- 
	delete_one(List1,H,List1Next),
	my_subtract(List1Next,T,List2).

%my_replace(+ListIn,+List1,+List2,?ListOut)
my_replace(ListIn,L1,L2,ListOut) :- 
	append(X,Y,ListIn),
	append(L1,Z,Y),
	append(L2,Z,Y1),
	append(X,Y1,ListOut), !.
my_replace(_,_,_,_) :- write('my_replace failed!'), nl.


% This predicate finds the intersection between two lists, but the occurences
% are added only once.
my_intersection([],_,[]) :- !.
my_intersection([H1|T1],List2,[H1|T]) :- 
	member(H1,List2), !,
	my_intersection(T1,List2,T).
my_intersection([_|T1],List2,T) :- 
	my_intersection(T1,List2,T).

my_union([],L,L) :- !.
my_union([H1|T1],List2,Union) :- 
	member(H1,List2), !,
	my_union(T1,List2,Union).
my_union([H1|T1],List2,[H1|T]) :- 
	my_union(T1,List2,T).

nu_intersection([],_,[]) :- !.
nu_intersection([H1|T1],List2,[H1|T]) :- 
	nu_member(H1,List2), !,
	nu_intersection(T1,List2,T).
nu_intersection([_|T1],List2,T) :- 
	nu_intersection(T1,List2,T).

% This predicate tests if two lists have the same elements (same sets) regardless
% the order. It does so by iterating the first set and deleting its head from the
% second list. When both empty at the same time -> success!
same_set([],[]) :- !.
same_set([H1|T1],L2) :- 
	delete_one_fail(L2,H1,NextL2),
	same_set(T1,NextL2).

% This predicate checks if a set S belongs to a list of sets
member_set(S,[H|_]) :- 
	same_set(S,H).
member_set(S,[_|T]) :- 
	member_set(S,T).

% This predicate deletes an element from a list only the first
% time the element is found.
delete_one([],_,[]) :- !.
delete_one([H|T],H,T) :- !.
delete_one([H|T],E,[H|T1]) :- 
	delete_one(T,E,T1).

delete_all([],_,[],[]) :- !.
delete_all([H|T],X,T1,[H|T2]) :- 
	subsumes_chk(X,H),
	delete_all(T,X,T1,T2).
delete_all([H|T],X,[H|T1],T2) :- 
	delete_all(T,X,T1,T2).

% This predicate deletes all occurences of an element from a list.
% (no unification at all!)
nu_delete([],_,[]) :- !.
nu_delete([H|T],E,T1) :- 
	\+ ( \+ (E=H)), !,
	nu_delete(T,E,T1).
nu_delete([H|T],E,[H|T1]) :- 
	nu_delete(T,E,T1).

nu_member(E,[H|_]) :- 
	\+ (\+ (E=H)), !.

nu_member(E,[_|T]) :- 
	nu_member(E,T).

reverse_append([],L,L) :- !.
reverse_append([H1|T1],L2,L) :- 
	reverse_append(T1,[H1|L2],L).


tidy_production_rules(CSIn,CSOut) :- 
	tidy_production_rules(CSIn,[],CSOut).

tidy_production_rules([],ConflictSetOut,ConflictSetOut) :- !.
tidy_production_rules([RuleOID-negative(Params)|RestIn],ConflictSetTmp,ConflictSetOut) :- !,
	delete(RestIn,RuleOID-positive(Params),RestNext),
	tidy_production_rules(RestNext,ConflictSetTmp,ConflictSetOut).
tidy_production_rules([RuleOID-positive(Params)|Tail],ConflictSetTmp,ConflictSetOut) :- !,
	delete(Tail,RuleOID-positive(Params),NextTail),
	tidy_production_rules(NextTail,[RuleOID-positive(Params)|ConflictSetTmp],ConflictSetOut).
tidy_production_rules([RuleOID-unique(positive(Params))|Tail],ConflictSetTmp,ConflictSetOut) :- !,
	delete(Tail,RuleOID-unique(positive(_)),NextTail),
	tidy_production_rules(NextTail,[RuleOID-positive(Params)|ConflictSetTmp],ConflictSetOut).
% The following is for rule instantiations that are not positive or negative
% ususally these are simple events activating production rules
tidy_production_rules([_|Tail],ConflictSetTmp,ConflictSetOut) :- !,
	tidy_production_rules(Tail,ConflictSetTmp,ConflictSetOut).


gather_all([],_,[],[]) :- !.
gather_all([H|T],X,T1,[Token|T2]) :- 
	X = RuleOID-Params1,
	H = RuleOID-Params,
	Params1 =.. [PosNeg,_Token1],
	Params =.. [PosNeg,Token], !,
	gather_all(T,X,T1,T2).
gather_all([H|T],X,[H|T1],T2) :- 
	gather_all(T,X,T1,T2).

make_set([],[]) :- !.
make_set([RuleOID-Params|Tail],[RuleOID-NewParams|ConflictSetTail]) :- 
	Params =.. [PosNeg,Token],
	SParams =.. [PosNeg,_],
	gather_all(Tail,RuleOID-SParams,NewTail,Deleted),
	NewParams =.. [PosNeg,[Token|Deleted]],
	make_set(NewTail,ConflictSetTail).

make_tuple([],[]) :- !.
make_tuple([RuleOID-ParamsSet|Tail],FinalList) :- 
	ParamsSet =.. [PosNeg,TokenSet],
	split_all(RuleOID,PosNeg,TokenSet,InstList),
	make_tuple(Tail,Tail1),
	append(InstList,Tail1,FinalList).

split_all(_,_,[],[]) :- !.
split_all(RuleOID,PosNeg,[Token|Tail1],[RuleOID-Param|Tail2]) :- 
	Param =.. [PosNeg,Token],
	split_all(RuleOID,PosNeg,Tail1,Tail2).
	

% This predicate deletes an element from a list only the first
% time the element is found. If the element does not exist it fails!
delete_one_fail([H|T],H,T) :- !.
delete_one_fail([H|T],E,[H|T1]) :- 
	delete_one_fail(T,E,T1).

% This predicate replaces an element found in a list with another element,
% giving another list.
replace_element([],_,_,[]) :- !.
replace_element([H|T],H,H1,[H1|T]) :- !.
replace_element([H|T1],Old,New,[H|T2]) :- 
	replace_element(T1,Old,New,T2).

% This predicate replaces all the elements in the second list, found in the first list 
% with another element, giving another list.
replace_last_element([],_,_,[]) :- !.
replace_last_element([H|T],[H],H1,[H1|T]) :- !.
replace_last_element([H|T],[H|T1],H1,L2) :- !,
	replace_last_element(T,T1,H1,L2).
replace_last_element([H|T1],Olds,New,[H|T2]) :- 
	replace_last_element(T1,Olds,New,T2).


find_min(List,Min) :- 
	List = [H|T],
	find_min(T,H,Min).

find_min([],Min,Min) :- !.
find_min([H|T],TempMin,Min) :- 
	H < TempMin, !,
	find_min(T,H,Min).
find_min([_|T],TempMin,Min) :- 
	find_min(T,TempMin,Min).

find_max(List,Min) :- 
	List = [H|T],
	find_max(T,H,Min).

find_max([],Max,Max) :- !.
find_max([H|T],TempMax,Max) :- 
	H > TempMax, !,
	find_max(T,H,Max).
find_max([_|T],TempMax,Max) :- 
	find_max(T,TempMax,Max).

sum_list(List,Sum) :- 
	sum_list(List,0,Sum).

sum_list([],Sum,Sum) :- !.
sum_list([H|T],Temp,Sum) :- 
	Next is Temp + H,
	sum_list(T,Next,Sum).


% This predicate replaces the Nth argument of OldTerm with Arg
% and returns it in NewTerm
my_setarg(N,OldTerm,Arg,NewTerm) :- 
	functor(OldTerm,Functor,Arity),
	N =< Arity,
	OldTerm =.. [Functor|Arguments],
	my_setarg_aux(N,Arguments,Arg,NewArguments),
	NewTerm =.. [Functor|NewArguments].

my_setarg_aux(1,[_OldArg|RestOld],Arg,[Arg|RestOld]) :- !.
my_setarg_aux(N,[OldArg|RestOld],Arg,[OldArg|RestNew]) :- 
	N1 is N - 1,
	my_setarg_aux(N1,RestOld,Arg,RestNew).

% This predicate takes a string and returns a list of variables names occurring
% in it.
var_list(String,List) :- 
	append(String,[46],String1),
	open_chars_stream(String1,Stream),
	read_term(Stream,_,[variable_names(VarList)]),
	close(Stream),
	var_names(VarList,List).

var_names([],[]) :- !.
var_names([Name=_|Tail],[Name|Tail1]) :- 
	var_names(Tail,Tail1).

% This predicate takes a list of variable names and returns a token
% of them in the format [...,var(VarName,VarValue),...]
var_token([],[]).
var_token([VarName=VarVal|Tail],[var(VarName,[VarVal])|TailVar]) :- 
	var_token(Tail,TailVar).

% This predicate unifies a list of variables (along with their names) with the vars
% found in token
unify_vars_in_token([],_) :- !.
unify_vars_in_token([VarName=Var|T1],Token) :- 
	member(var(VarName,[Var]),Token), !,
	unify_vars_in_token(T1,Token).
unify_vars_in_token([_VarName=_Var|T1],Token) :- 
	unify_vars_in_token(T1,Token).



% This predicate generates a new symbol every time is run, by concatenating
% an increasing integer at the end of the atom Root
gensym(Root,Atom) :-
	get_num(Root,Number),
	atom_chars(Root,RootString),
	number_chars(Number,NumberString),  %PIPE
	append(RootString,NumberString,FinalString), %PIPE
	atom_chars(Atom,FinalString).

:- dynamic current_count/2.

get_num(Root,Number) :- 
	retract(current_count(Root,PreviousNumber)), !,
	Number is PreviousNumber + 1,
	assert(current_count(Root,Number)).
get_num(Root,1) :- 
	asserta(current_count(Root,1)).

% This predicate generates a new variable with root 'XX' for
% the parsing of DEVICE rules. These variables are used for
% intermediate variables for transforming path expressions
% into simple ones.
gen_new_var(Vars,V) :- 
	gensym('XX',V),
	\+ (member(V,Vars)), !.
gen_new_var(Vars,V) :- 
	gen_new_var(Vars,V).

% This predicate resets the counter for the variables 'XX'
% each time a new rule is parsed.
reset_var_gen :- 
	retract(current_count('XX',_)), !.
reset_var_gen.

% This predicate checks whether its argument is a list.
my_is_list([]) :- !.
my_is_list([_|_]).

% This predicate transforms the parameters of mm_ms_event occurences 
% in EXACT to primitive event occurrences that are needed for DEVICE
transform_parameters(_DEventOID,occurrence_tuple(EventOID,Params),_,occurrence_tuple(EventOID,Params)) :- 
	\+ (my_instance_of(EventOID,mm_inheritable_ms_event)), !.
transform_parameters(_DEventOID,occurrence_tuple(EventOID,Params),_,occurrence_tuple(EventOID,Params)) :- 
	Params =.. [PosNeg,_Token],
	member(PosNeg,[positive,negative]), !.
transform_parameters(DEventOID,occurrence_tuple(EventOID,[_,OID,[Val]]),PosNeg,
		               occurrence_tuple(EventOID,ParamTuple)) :- 
	get_out_token(token(L1,L2,[var(OID_var,_),var(Var,_)])) => EventOID, 
	(member(DEventOID,L1) -> true; member(DEventOID,L2)),
	ParamTuple =.. [PosNeg,[var(OID_var,[OID]),var(Var,[Val])]].
transform_parameters(DEventOID,occurrence_tuple(EventOID,[_,OID,[Val]]),PosNeg,
		               occurrence_tuple(EventOID,ParamTuple)) :- 
	get_out_token(token(L1,L2,[var(OID_var,_),var(Var1,_),var(Var2,_)])) => EventOID, 
	(member(DEventOID,L1) -> true; member(DEventOID,L2)),
	ParamTuple =.. [PosNeg,[var(OID_var,[OID]),var(Var1,[Val]),var(Var2,[Val])]].
transform_parameters(DEventOID,occurrence_tuple(EventOID,[_,_OID,[Val]]),PosNeg,
		               occurrence_tuple(EventOID,ParamTuple)) :- 
	get_out_token(token(L1,L2,[var(Var,_)])) => EventOID, 
	(member(DEventOID,L1) -> true; member(DEventOID,L2)),
	ParamTuple =.. [PosNeg,[var(Var,[Val])]].
transform_parameters(DEventOID,occurrence_tuple(EventOID,[new,_Class,[OID|_]]),PosNeg,
		               occurrence_tuple(EventOID,ParamTuple)) :- 
	get_out_token(token(L1,L2,[var(OID_var,_)])) => EventOID, 
	(member(DEventOID,L1) -> true; member(DEventOID,L2)),
	ParamTuple =.. [PosNeg,[var(OID_var,[OID])]].



% This predicate transforms an 'update' event to 'delete' and 'put' events
transform_update_parameters([PEventOID,DEventOID],AEventOID,
		     	    occurrence_tuple(EventOID,[_,OID,[OldVal,NewVal]]),[Ptuple,Dtuple]) :- 
	get_out_token(token(L1,_L2,_)) => EventOID, 
	member([PEventOID,DEventOID],L1), !,
	setof(E,(get_activated_events(E) => PEventOID,
	         get_de_activated_events(E) => DEventOID), Es),
	member(AEventOID,Es),	
	transform_parameters(AEventOID,occurrence_tuple(PEventOID,[_,OID,[NewVal]]),positive,Ptuple),
	transform_parameters(AEventOID,occurrence_tuple(DEventOID,[_,OID,[OldVal]]),negative,Dtuple),
	similar_tuples(Ptuple,Dtuple).

transform_update_parameters_rule([EventOID,AntiEventOID],RuleOID,
		     	    occurrence_tuple(UEventOID,[_,OID,[OldVal,NewVal]]),[Ptuple,Ntuple]) :- 
	get_out_token(token(L1,_L2,_)) => UEventOID, 
	member([EventOID,AntiEventOID],L1), !,
	setof(R,(get_activated_rules(R) => EventOID,
		  get_de_activated_rules(R) => AntiEventOID), Rs),
	member(RuleOID,Rs),
	transform_parameters(RuleOID,occurrence_tuple(EventOID,[_,OID,[NewVal]]),positive,Ptuple),
	transform_parameters(RuleOID,occurrence_tuple(AntiEventOID,[_,OID,[OldVal]]),negative,Ntuple),
	similar_tuples(Ptuple,Ntuple).

similar_tuples(occurrence_tuple(_,Tuple1),occurrence_tuple(_,Tuple2)) :-
	Tuple1 =.. [_,T1],
	Tuple2 =.. [_,T2],
	similar_tuples_aux(T1,T2).
	
similar_tuples_aux([],[]) :- !.
similar_tuples_aux([var(Var,_)|T1],[var(Var,_)|T2]) :- 
	similar_tuples_aux(T1,T2).

% This predicate creates an executable Prolog compound goal that
% serves as the predicate check of logical events
create_logical_predicate(single([pred(Op,Const)]),Val,Pred) :- !,
	Pred =.. [Op,Val,Const].
create_logical_predicate(and(Preds),Val,Pred) :- !,
	create_complex_predicate(',',Preds,Val,Pred).
create_logical_predicate(or(Preds),Val,Pred) :- !,
	create_complex_predicate(';',Preds,Val,Pred).

create_complex_predicate(_,[pred(Op,Const)],Val,Pred) :- 
	Pred =.. [Op,Val,Const].
create_complex_predicate(Connective,[pred(Op,Const)|Rest],Val,Pred) :- 
	create_complex_predicate(Connective,Rest,Val,RestPred),
	First =.. [Op,Val,Const],
	Pred =.. [Connective,First,RestPred].

% This predicate creates the join predicates
create_join_predicates(LR_OP,RL_OP,LVal,RVal,LR_pred,RL_pred) :- 
	create_one_predicate(LR_OP,LVal,RVal,LR_pred),
	create_one_predicate(RL_OP,RVal,LVal,RL_pred).

create_one_predicate(OPs,LVals,RVals,Pred) :- 
	length(OPs,L),
	length(RVals,L), !,
	create_one_predicate_aux1(OPs,LVals,RVals,Pred).
create_one_predicate(OPs,LVals,RVals,Pred) :- 
	length(OPs,L),
	length(LVals,L), !,
	create_one_predicate_aux3(OPs,LVals,RVals,Pred).

create_one_predicate_aux1(OPs,[LVal],RVals,Pred) :- !,
	create_one_predicate_aux2(OPs,LVal,RVals,Pred).
create_one_predicate_aux1(OPs,[LVal|TLVals],RVals,Pred) :- 
	create_one_predicate_aux2(OPs,LVal,RVals,PredH),
	create_one_predicate_aux1(OPs,TLVals,RVals,PredT),
	Pred = (PredH, PredT).

create_one_predicate_aux2([LR_OP],LVal,[RVal],LR_pred) :- !,
	LR_pred =.. [LR_OP,LVal,RVal].
create_one_predicate_aux2([LR_OP|T_op],LVal,[RVal|TRVal],LR_pred) :- 
	LR_predH =.. [LR_OP,LVal,RVal],
	create_one_predicate_aux2(T_op,LVal,TRVal,LR_predT),
	LR_pred = (LR_predH, LR_predT).

create_one_predicate_aux3(OPs,LVals,[RVal],Pred) :- !,
	create_one_predicate_aux4(OPs,LVals,RVal,Pred).
create_one_predicate_aux3(OPs,LVals,[RVal|TRVals],Pred) :- 
	create_one_predicate_aux4(OPs,LVals,RVal,PredH),
	create_one_predicate_aux3(OPs,LVals,TRVals,PredT),
	Pred = (PredH, PredT).

create_one_predicate_aux4([LR_OP],[LVal],RVal,LR_pred) :- !,
	LR_pred =.. [LR_OP,LVal,RVal].
create_one_predicate_aux4([LR_OP|T_op],[LVal|TLVal],RVal,LR_pred) :- 
	LR_predH =.. [LR_OP,LVal,RVal],
	create_one_predicate_aux4(T_op,TLVal,RVal,LR_predT),
	LR_pred = (LR_predH, LR_predT).


% This predicate creates a list of unbound vars from a list of operators.
create_value([],[]) :- !.
create_value([_|T],[_|T1]) :- create_value(T,T1).

% This predicate creates the comparison operator for the two-input DEVICE events
create_comparison(Op1,['='],Op1,2) :- !.
create_comparison(['='],Op2,Oper,1) :- !,
	inverse_operator(Op2,Oper).
create_comparison([Op1],_,[Op1],3) :- !,
	Op1 \== '=',
	Op1 \== '\=='.
create_comparison(Op1,_,Op1,3).

inverse_operator([],[]) :- !.
inverse_operator([X|T],[Y|T1]) :- !,
	inverse_operator(X,Y),
	inverse_operator(T,T1).
inverse_operator(X,Y) :- inverse_oper(X,Y), !.
inverse_operator(X,Y) :- inverse_oper(Y,X), !.

inverse_oper(=,=).
inverse_oper(\==,\==).
inverse_oper(>,<).
inverse_oper(>=,=<).

create_out_value(Event_type,_,LVal,_,LVal) :- 
	my_subclass_of(Event_type,mm_negative_event), !.
create_out_value(_,1,LVal,_RVal,LVal) :- !.
create_out_value(_,2,_LVal,RVal,RVal) :- !.
create_out_value(_,3,LVal,RVal,Out_Val) :- 
	append(LVal,RVal,Out_Val).

deletion_optimization(['='],['='],on) :- !.
deletion_optimization(_,_,off).

% This predicate takes a list of positive and negative event occurrences (with the tag positive and negative)
% and splits them into two lists accordingly.
transform_to_normal_occs([],[],[]) :- !.
transform_to_normal_occs([occurrence_tuple(Event,positive(Params))|Rest],
	                 [occurrence_tuple(Event,Params)|T1],T2) :- !,
	transform_to_normal_occs(Rest,T1,T2).
transform_to_normal_occs([occurrence_tuple(Event,negative(Params))|Rest],T1,
	                 [occurrence_tuple(Event,Params)|T2]) :- !,
	transform_to_normal_occs(Rest,T1,T2).
transform_to_normal_occs([occurrence_tuple(Event,Params)|Rest],
	                 [occurrence_tuple(Event,Params)|T1],T2) :- 
	transform_to_normal_occs(Rest,T1,T2).

% This predicate performs the lr or rl join between all the vars in a two-input event
check_all(_,[],LToken,RToken,OutToken) :- !,
	my_union(LToken,RToken,OutToken).
check_all(LR,[H|T],LToken,RToken,[var(Var,OutVal)|RestOut]) :- 
	H = var(Var,OutVal,LVal,RVal,LR_pred,RL_pred),
	delete_one(LToken,var(Var,LVal),NextLToken),
	delete_one(RToken,var(Var,RVal),NextRToken),
	(LR == lr ->
	    call(LR_pred);
	    call(RL_pred)),
	check_all(LR,T,NextLToken,NextRToken,RestOut).

% This predicate performs the lr or rl merging of the input token values
% between all the vars in a two-input event
merge_all(_,[],LToken,RToken,OutToken) :- !,
	append(LToken,RToken,OutToken).
merge_all(LR,[H|T],LToken,RToken,[var(Var,OUTVAL)|RestOut]) :- !,
	H = var(Var,_OutVal,LVal,RVal,_LR_pred,_RL_pred),
	delete_one(LToken,var(Var,LVal),NextLToken),
	delete_one(RToken,var(Var,RVal),NextRToken),
	(LR = lr -> 
		OUTVAL = LVal;
		OUTVAL = RVal),
	merge_all(LR,T,NextLToken,NextRToken,RestOut).



% This predicate discriminates between production rule occurrences and other rule types.
choose_production_rules([],[],[]) :- !.
choose_production_rules([RuleOID-Params|Rest1],[RuleOID-Params|Rest2],Rest3) :- 
	my_instance_of(RuleOID,mm_production_rule), !,
	choose_production_rules(Rest1,Rest2,Rest3).
choose_production_rules([Rule|Rest1],Rest2,[Rule|Rest3]) :- 
	choose_production_rules(Rest1,Rest2,Rest3).


% This predicate searches in the ordered_rule_list for the position that the new production
% rule should be added, according to the specificity priority criterion. Note that the position
% is indicated by the rule OID before which the new rule is added. This is used in the priority
% function of production_rule!
find_before_rule(_,[],[]) :- !.
find_before_rule(Total_priority,[Before-_|_],[Before]) :- 
	get_event_number(N) => Before,
	(get_control(yes) => Before -> 
		CP = 0;
		CP = 1),
	get_priority(P) => Before,
	Total is 1000 * CP + 100 * P + N,
	Total =< Total_priority, !.
find_before_rule(No_of_events,[_|Tail],Before) :- 
	find_before_rule(No_of_events,Tail,Before).

% Debugging hacks!
stop_here.
start_here.

% This predicate transforms a list of vars from the parser's to the compiler's format
transform_vars([],[]) :- !.
transform_vars([V|Rest1],[var(V,[_])|Rest2]) :- 
	transform_vars(Rest1,Rest2).


% This predicate restores the natural ordering of vars and events in normal Var_list
fix_list(List1,List2) :- fix_list(List1,List2,[]), !.

fix_list([],List2,List2) :- !.
fix_list([var(Var,REvent_list,Action)|T1],List2,Temp) :- 
	reverse(REvent_list,Event_list),
	fix_list(T1,List2,[var(Var,Event_list,Action)|Temp]).

% This predicate extracts the negative events from nested negative inter-object patterns!
extract_negative_events([],[]) :- !.
extract_negative_events([var(_,EVs,_)|Rest],AllNegs) :- 
	findall(Neg,member(ev(Neg,_,_,_),EVs),NegEvents),
	extract_negative_events(Rest,Rest1),
	append(NegEvents,Rest1,AllNegs).

% This predicate extracts the variable on which the join inside a nested negative pattern is performed
% Join on an OID variable (half normal)
join_on_which_variable(EV1,EV2,AEV1,AEV2,Var_listIn,OID_Var_listIn,
                       Var,OP1,Left_in_token,OP2,Right_in_token) :- 
	member(var(Var,Event_list1,_),Var_listIn),
	member(ev(EV1,AEV1,OP1,Left_in_token),Event_list1),
	member(var(Var,Event_list2,_),OID_Var_listIn),
	member(ev(EV2,AEV2,OP2,Right_in_token),Event_list2), !.
% Join on a normal variable
join_on_which_variable(EV1,EV2,AEV1,AEV2,Var_listIn,_,Var,OP1,Left_in_token,OP2,Right_in_token) :- 
	member(var(Var,Event_list,_),Var_listIn),
	my_subset([ev(EV1,AEV1,OP1,Left_in_token),ev(EV2,AEV2,OP2,Right_in_token)],Event_list), !.
% No join but cartesian. Return var Var
join_on_which_variable(EV1,EV2,AEV1,AEV2,Var_listIn,_OID_Var_listIn,
                       _,OP1,Left_in_token,OP2,Right_in_token) :- 
	member(var(_Var1,Event_list1,_),Var_listIn),
	member(ev(EV1,AEV1,OP1,Left_in_token),Event_list1),
	member(var(_Var2,Event_list2,_),Var_listIn),
	member(ev(EV2,AEV2,OP2,Right_in_token),Event_list2), !.


% This predicate checks if some of the rule instantiations in the calculated conflict set 
% for the production rules have already been selected for activation in the current cycle.
% If these are not removed, then the rule manager cycles, because the calculate_conflict_set
% method would always return the same and the same rules. This happens because the rule
% firing is done by an ECA rule itself!
check_if_already_activated([],Occs,Occs) :- !.
check_if_already_activated(_,[],[]) :- !.
check_if_already_activated(Act_set,[Occ|Rest],Rest1) :- 
	member(Occ,Act_set), !,
	check_if_already_activated(Act_set,Rest,Rest1).
check_if_already_activated(Act_set,[Occ|Rest],[Occ|Rest1]) :- 
	check_if_already_activated(Act_set,Rest,Rest1).

% This predicate collects together all rule instantiations that are going to be activated!
% This is used for recording activated rules.
find_activated_inst([],[]) :- !.
find_activated_inst([occurrence_tuple(_,[none,_,[Partial_confict_set]])|RestOccs],TotalInsts) :- 
	find_activated_inst(RestOccs,RestInsts),
	append(Partial_confict_set,RestInsts,TotalInsts).

% This predicates signals the beginning of production rule processing. Note that both
% main-memory and disk-based production rules are of the user-invoked coupling mode,
% because of the nested evaluation of production rules. (See below the next_cycle pred).
% 
fc :- 
	fc => mm_production_rule_manager.

stop :- 
	message_recipient(Rule),
	instance_of(Rule,RuleClass),
	stop => RuleClass.


switch_task(Task) :-
	message_recipient(Rule),
	instance_of(Rule,RuleClass),
	(get_task(Current) => RuleClass -> 
		update_task([Current,Task]) => RuleClass;
		put_task([Task]) => RuleClass),
	(get_spy_tasks(on) => RuleClass ->
		write_list(['Switching task from',Current,'to',Task]);
		true).

:- dynamic '$spy_loader'/1.

spy_loader :- 
	(clause('$spy_loader'(on),true) ->
		(write('Loader is already being spied!'), nl);
		(write('Loader is being spied!'), nl,
		 (retract('$spy_loader'(_)) -> true; true),
		 assertz('$spy_loader'(on)))).
no_spy_loader :- 
	(clause('$spy_loader'(off),true) ->
		(write('Loader is not being spied!'), nl);
		(write('Loader is no longer being spied!'), nl,
		 (retract('$spy_loader'(_)) -> true; true),
		 assertz('$spy_loader'(off)))).

spy_cs :- 
	spy => mm_production_rule_manager.

nospy_cs :- 
	nospy => mm_production_rule_manager.


store_diff_cs(DiffCS,RuleClass) :- 
	(get_previous_cs(Prev) => RuleClass ->
		(append(DiffCS,Prev,Next),
		 update_previous_cs([Prev,Next]) => RuleClass);
		put_previous_cs([DiffCS]) => RuleClass).


spy_tasks :- 
	(get_spy_tasks(A) => mm_production_rule ->
	 	(A \== on ->
			(update_spy_tasks([A,on]) => mm_production_rule,
		 	 write('Now spying rule modules'), nl);
			write('Rule modules are already being spied!'), nl);
		(put_spy_tasks([on]) => mm_production_rule,
		 write('Now spying rule modules'), nl)).

no_spy_tasks :- 
	(get_spy_tasks(A) => mm_production_rule ->
	 	(A \== off ->
			(update_spy_tasks([A,off]) => mm_production_rule,
		 	 write('Rule modules are no more spied.'), nl);
			write('Rule modules are not being spied!'), nl);
		(put_spy_tasks([off]) => mm_production_rule,
		 write('Rule modules are not spied.'), nl)).


spy_events([]) :- !.
spy_events([H|T]) :- 
	spy => H,
	spy_events(T).

no_spy_events([]) :- !.
no_spy_events([H|T]) :- 
	no_spy => H,
	no_spy_events(T).

display_net(Rule) :- 
	((atom(Rule),
	  get_by_name([Rule],RuleOID) => mm_production_rule) ->
		display_net => RuleOID;
		write_list(['DEVICE error! Rule with name:',Rule,', does not exist!'])).

spy_net(Rule) :- 
	((atom(Rule),
	  get_by_name([Rule],RuleOID) => mm_production_rule) ->
		spy_net => RuleOID;
		write_list(['DEVICE error! Rule with name:',Rule,', does not exist!'])).

spy_net(Rule,Depth) :- 
	((atom(Rule),
	  get_by_name([Rule],RuleOID) => mm_production_rule) ->
		spy_to_depth([Depth]) => RuleOID;
		write_list(['DEVICE error! Rule with name:',Rule,', does not exist!'])).


spy_nested_events(_,_,[]) :- !.
spy_nested_events(D,D,_) :- !.
spy_nested_events(Current,Depth,[H|T]) :- 
	my_is_list(H), !,
	Next is Current + 1,
	spy_nested_events(Next,Depth,H),
	spy_nested_events(Current,Depth,T).
spy_nested_events(Current,Depth,[H|T]) :- 
	spy => H,
	spy_nested_events(Current,Depth,T).


no_spy_net(Rule) :- 
	((atom(Rule),
	  get_by_name([Rule],RuleOID) => mm_production_rule) ->
		no_spy_net => RuleOID;
		write_list(['DEVICE error! Rule with name:',Rule,', does not exist!'])).

no_spy_net(Rule,Depth) :- 
	((atom(Rule),
	  get_by_name([Rule],RuleOID) => mm_production_rule) ->
		no_spy_to_depth([Depth]) => RuleOID;
		write_list(['DEVICE error! Rule with name:',Rule,', does not exist!'])).

% This predicate print a nested list of events in displaced columns.
no_spy_nested_events(_,_,[]) :- !.
no_spy_nested_events(D,D,_) :- !.
no_spy_nested_events(Current,Depth,[H|T]) :- 
	my_is_list(H), !,
	Next is Current + 1,
	no_spy_nested_events(Next,Depth,H),
	no_spy_nested_events(Current,Depth,T).
no_spy_nested_events(Current,Depth,[H|T]) :- 
	no_spy => H,
	no_spy_nested_events(Current,Depth,T).

% This predicate is used from a rule to ask questions to the user
query([],Answer):- get(CharCode), atom_chars(Answer,[CharCode]).
query([X|Xs],Answer):-
	write(X),
	query(Xs,Answer).


% Theis predicate checks if the former rule manager shares the conflict set 
% with the latter
same_conflict_set(RuleManager,RuleManager) :- !.
same_conflict_set(RuleManager1,RuleManager) :- 
	get_common_cs(yes) => RuleManager1,
	get_is_a(RuleManager2) => RuleManager1,
	same_conflict_set(RuleManager2,RuleManager).


% This predicate print a nested list in displaced columns.
print_nested(_,[]) :- !.
print_nested(Tab,[H|T]) :- 
	my_is_list(H), !,
	NextTab is Tab + 5,
	print_nested(NextTab,H),
	print_nested(Tab,T).
print_nested(Tab,[H|T]) :- 
	tab(Tab), write(H), nl,
	print_nested(Tab,T).

% This predicate print a nested list of events in displaced columns.
print_nested_events(_,[]) :- !.
print_nested_events(Tab,[H|T]) :- 
	my_is_list(H), !,
	NextTab is Tab + 5,
	print_nested_events(NextTab,H),
	print_nested_events(Tab,T).
print_nested_events(Tab,[H|T]) :- 
	tab(Tab), write(H), 
	print_event_info(H),
	print_nested_events(Tab,T).

print_event_info(Event) :- 
	my_instance_of(Event,mm_production_rule), !, nl.
print_event_info(Event) :- 
	my_instance_of(Event,mm_two_input_event), !,
	(get_join(Join_var_list) => Event ->
		(collect_join_var_names(Join_var_list,Vars),
		 write('  --  '),
		 write_list_with_commas(Vars), nl);
		(write('  --  '),
		 write('Cartesian'), nl)).
print_event_info(Event) :- 
	my_instance_of(Event,mm_logical_event), !,
	get_predicate([_,Pred]) => Event,
	write('  --  '),
	write(Pred), nl.
print_event_info(Event) :- 
	my_instance_of(Event,mm_ms_event), !,
	get_active_method(Method) => Event,
	get_active_class(Class) => Event,
	write('  --  '),
	write_list([Method,at,Class]). 

	
collect_join_var_names([],[]) :- !.
collect_join_var_names([var(Var,_,_,_,_,_)|T],[Var|TVars]) :- 
	collect_join_var_names(T,TVars).

write_list_with_commas([]) :- !.
write_list_with_commas([H]) :- !,
	write(H).
write_list_with_commas([H|T]) :- 
	write(H), write(', '),
	write_list_with_commas(T).

% This predicate merges together the values carried from two (or more) token elements (for the same
% variable) in one element with two (a list with two) values.
% e.g. [var('S1',[100]),var('E1',[0#emp]),var('S1',[90])] -> [var('S1',[100,90]),var('E1',[0#emp])]
% The order is important!
merge_duplicates([],[]) :- !.
merge_duplicates([var(Var,Val1)|Rest],Final) :- 
	delete_one_fail(Rest,var(Var,Val2),Rest2), !,
	append(Val1,Val2,NewVal),
	merge_duplicates([var(Var,NewVal)|Rest2], Final).
merge_duplicates([var(Var,Val)|Rest],[var(Var,Val)|Rest1]) :- 
	merge_duplicates(Rest,Rest1).


get_event_type(Event_typeIn,Event_typeOut) :- 
	message_recipient(MR),
	(mm_name(MR,_) -> 
		concat_atoms(mm_,Event_typeIn,Event_typeOut);
		Event_typeOut = Event_typeIn).

my_instance_of(Instance,Class) :- 
	instance_of(Instance,C),
	path(C,Supers),
	memberchk(Class,[C|Supers]), !.

my_subclass_of(SubClass,Class) :- 
	path(SubClass,Supers),
	member(Class,[SubClass|Supers]), !.


collect_join_vars(_,[],[]) :- !.
collect_join_vars(LR,[HJVar|TJVar],[Pred|RestPreds]) :- 
	HJVar = var(_Var,_,_,_,LPred,RPred),
	(LR == lr ->
		Pred = LPred;
		Pred = RPred),
	collect_join_vars(LR,TJVar,RestPreds).

% Only the last (exhaustive) call can be used!
choose_goal(In_token,SToken,Join_var_list,Predicates,[[Guard,Call]],Call) :- !,
	Guard = occurrence_params(In_token,SToken,Join_var_list,Predicates).
choose_goal(In_token,SToken,Join_var_list,Predicates,[[Guard,Call]|_],Call) :- 
	Guard = (occurrence_params(In_token,SToken,Join_var_list,Predicates), CallCheck),
	call(CallCheck), !. % The guard is confirmed, cut alternative calls!
% The guard is not confirmed, try alternative calls!
choose_goal(In_token,SToken,Join_var_list,Predicates,[_|Rest],Call) :- 
	choose_goal(In_token,SToken,Join_var_list,Predicates,Rest,Call).

partially_bound(_,_,[],[]) :- !.
partially_bound(In_token,SToken,[H|T],[_=_|RestPreds]) :- 
	H = var(Var,_,_,_,_,_),
	member(var(Var,Val),In_token),
	member(var(Var,Val),SToken), !,		% STOKEN-access 
	partially_bound(In_token,SToken,T,RestPreds).
partially_bound(In_token,SToken,[_|T],[_|RestPreds]) :- 
	partially_bound(In_token,SToken,T,RestPreds).

%templ_to_token(STokenTemplate,SToken),
templ_to_token([],[]) :- !.
templ_to_token([var(Var,Val,_,_)|T1],[var(Var,Val)|T2]) :- 
	templ_to_token(T1,T2).


retrieve_vars(_,[],[],true) :- !.
retrieve_vars(MemoryToken,[var(Var,Val,true,_)|TokenTemplateRest],[var(Var,Val)|RestStored],Calls) :- !,
	member(var(Var,Val),MemoryToken),
	retrieve_vars(MemoryToken,TokenTemplateRest,RestStored,Calls).
retrieve_vars(MemoryToken,[var(Var,Val1,false,_)|TokenTemplateRest],Stored,Calls) :- 
	member(var(Var,Val2),MemoryToken),
	all_instantiated(Val2), !,
	Val1 = Val2,
	retrieve_vars(MemoryToken,TokenTemplateRest,Stored,Calls).
retrieve_vars(MemoryToken,[var(Var,Val,false,Code)|TokenTemplateRest],Stored,(CodeOut,Calls)) :- 
	member(var(Var,Val),MemoryToken),
	Code = (var(OIDVar,OIDVal), CodeOut),
	member(var(OIDVar,OIDVal),MemoryToken),
	retrieve_vars(MemoryToken,TokenTemplateRest,Stored,Calls).

collect_vars([],[],[],true) :- !.
collect_vars([var(Var,Val)|RestMemoryToken],
		[var(Var,Val,true,_)|TokenTemplateRest],[var(Var,Val)|RestStored],Calls) :- !,
	collect_vars(RestMemoryToken,TokenTemplateRest,RestStored,Calls).
collect_vars([var(Var,Val),var(OIDVar,OIDVal)|RestMemoryToken],
		[var(Var,Val,false,Code)|TokenTemplateRest],Stored,(CodeOut,Calls)) :- 
	Code = (var(OIDVar,OIDVal), CodeOut),
	collect_vars(RestMemoryToken,TokenTemplateRest,Stored,Calls).

retrieve_stored_vars(_,[],[]) :- !.
retrieve_stored_vars(MemoryToken,[var(Var,Val,true,_)|TokenTemplateRest],[var(Var,Val)|RestStored]) :- !,
	member(var(Var,Val),MemoryToken),
	retrieve_stored_vars(MemoryToken,TokenTemplateRest,RestStored).
retrieve_stored_vars(MemoryToken,[var(_Var,_Val,false,_)|TokenTemplateRest],Stored) :- 
	retrieve_stored_vars(MemoryToken,TokenTemplateRest,Stored).

all_instantiated([]) :- !.
all_instantiated([H|T]) :- 
	nonvar(H),
	all_instantiated(T).

is_join_var_not_instantiated(Event,Token) :- 
		get_join(Join_var_list) => Event,
		is_not_fully_instantiated(Join_var_list,Token).



is_not_fully_instantiated([],_) :- !, fail.
is_not_fully_instantiated([var(Var,_,_,_,_,_)|_],Token) :- 
	member(var(Var,[Val]),Token),
	var(Val), !.
is_not_fully_instantiated([var(Var,_,_,_,_,_)|_],Token) :- 
	member(var(Var,[Val1,Val2]),Token),
	var(Val1), var(Val2), !.
is_not_fully_instantiated([_|Tail],Token) :- 
	is_not_fully_instantiated(Tail,Token), !.

is_it_total(ConstituentEvent) :- 
	my_instance_of(ConstituentEvent,mm_inheritable_ms_event),
	get_active_method(Method) => ConstituentEvent,
	atom_chars(Method,MethodS),
	append("delete_",AttributeS,MethodS),  %PIPE
	atom_chars(Attribute,AttributeS),
	get_active_class(Class) => ConstituentEvent,
	get_slot_desc(slot_desc_tuple(Attribute,Class,_,_,total,_)) => Class.

find_cardinality(Attribute,Class,S) :- 
	path(Class,Supers),
	member(X,[Class|Supers]),
	get_slot_desc(slot_desc_tuple(Attribute,X,local,S,_,_)) => X,
	!.

find_update_event(N1#mm_inheritable_ms_event,N2#mm_inheritable_ms_event) :- 
	N2 is N1 + 1.

write_list_in_lines([]) :-  nl, !.
write_list_in_lines([H|T]) :- 
	write(H), nl,
	write_list_in_lines(T).
	


/******************************************************/
/** The following pred were not needed in SICStus 3.5 */
/******************************************************/
concat_atoms(X,Y,Z):-
	nonvar(X), nonvar(Y), !,
	atom_chars(X,XL),
	atom_chars(Y,YL),
	append(XL,YL,ZL),
	atom_chars(Z,ZL).

concat_atoms(X,Y,Z):-
	nonvar(Z), !,
	atom_chars(Z,ZL),
	append(XL,YL,ZL),
	atom_chars(X,XL),
	atom_chars(Y,YL).

