%****************************************************************************************

%*************************************************************************
% FILE:		device_rule_manager.pl
% SUBJECT:	It is the definition of metaclasses, classes, auxiliar
%		predicates and rules that compound the rule manager for 
%		the DEVICE system
% AUTHOR:	NICK BASSILIADES
% DATE:		October-December 1995
%*************************************************************************

:- new([meta_production_rule_manager,[
    is_a([meta_rule_manager]),

	% The following iterates over all autonomous conflict set rule managers
        % sub-classes of mm_production_rule and send scheduling and de-scheduling
	% messages. Using this way there is no need to re-define this method
	% every time a new extension to DEVICE is done! (e.g. deductive rules)

    method((schedule_all(global,[],[plog],[],[Occs]) :-
	((get_by_common_cs([no],RuleManager) => self,
	  scheduling([Occs]) => RuleManager,
	  fail);
	 true)

     )),

    method((fc(global,[],[],[],[]) :-
	((message_recipient(MetaRuleManager),
	  get_by_common_cs([no],RuleManager) => MetaRuleManager,
	  fc => RuleManager,
	  fail);
	 true)
     )),


    method((spy(global,[],[],[],[]) :-
	((message_recipient(MetaRuleManager),
	  get(A) => MetaRuleManager,
	  spy => A,
	  fail);
	 true)
     )),

    method((no_spy(global,[],[],[],[]) :-
	((message_recipient(MetaRuleManager),
	  get(A) => MetaRuleManager,
	  no_spy => A,
	  fail);
	 true)
     ))

]]) => meta_class.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
%	metaclass MM_PRODUCTION_RULE_MANAGER
% The instance of 'mm_production_rule_manager' is the production rule manager.
% 
:- new([mm_production_rule_manager,[
    is_a([mm_standard_rule_manager]),

%   ----------------------------------------------------------------------
    slot(slot_tuple(conflict_set,  global,single,optional,plog)),
	% This slot holds a yes or no and defines the conflict set behaviour
	% A 'yes' indicates that the rule instantiations of the rule manager
	% are hold in the conflict set of the most general super-class of the kind
	% which is 'mm_production_rule'. A 'no' indicates that the rule instantiations
	% are hold in a separate local conflict set of the rule manager.
	% (e.g. deductive rules) 
    slot(slot_tuple(common_cs,  global,single,total,string)),

    % This slot stores the scheduling event (type mm_system_abstract_event)
    % created by the 'scheduling_rule_creation' method (of this class). There
    % is a different scheduling event attached to each rule manager, and the
    % the rules activated for it (system_rule) manage the conflict set of 
    % rules of the particular rule manager (i.e. mm_generic_rule).
    %
    % In the case of mm_system_rule, this slot is empty as the rules of this
    % rule manager are managed (fired) directly in the execution of the
    % method 'scheduling'
    %
    slot(slot_tuple(refractoriness,     global,single,mandatory,string)),
    slot(slot_tuple(recency,            global,single,mandatory,string)),
    slot(slot_tuple(specificity,        global,single,mandatory,string)),
    slot(slot_tuple(task,               global,single,optional,string)),
    slot(slot_tuple(status,             global,single,mandatory,string)),
    slot(slot_tuple(spy_conflict_set,   global,single,mandatory,string)),
    slot(slot_tuple(spy_tasks,   	global,single,mandatory,string)),
    slot(slot_tuple(previous_cs,   	global,single,optional,plog)),
    slot(slot_tuple(cycle_no,   	global,single,optional,integer)),


%   +--------------------------------------------------------------------+
%   |	method NEW							 |
%   |	The execution of this method creates a new individual rule such as 
%   |	1#mm_production_rule:
%   |
    method((new(global,[],[plog,plog],[],[RuleOid,Atts]) :-
      ( message_recipient(RuleClass),
	delete(Atts,rule([Rule1]),Atts1),
	(nonvar(Rule1) -> 
	        ((atom(Rule1) -> RuleAtom1=Rule1; atom_chars(RuleAtom1,Rule1)),
		 (member(module([Module]),Atts1) ->
			(concat_atoms('if ',RestAtom,RuleAtom1) -> 
			 	(concat_atoms('mm_production_rule(task=',Module,S1),
			 	 concat_atoms(S1,') and ',AddString),
			 	 concat_atoms('if ',AddString,NewRule1),
			 	 concat_atoms(NewRule1,RestAtom,RuleAtom));
				(concat_atoms('if-true ',RestAtom,RuleAtom1),
				 concat_atoms('mm_production_rule(task=',Module,S1),
			 	 concat_atoms(S1,') ',AddString),
			 	 concat_atoms('if ',AddString,NewRule1),
			 	 concat_atoms(NewRule1,RestAtom,RuleAtom)));
			RuleAtom = RuleAtom1),
		 (member(control([_]),Atts1) ->
			Atts2 = Atts1;
			Atts2 = [control([no])|Atts1]),
		 delete(Atts2,priority([Priority]),Atts3),
		 (nonvar(Priority) ->
			(Priority > 9 -> 
				Priority1 = 9;
				(Prority < 0 -> 
					Priority1 = 0;
					Priority1 = Priority));
			Priority1 = 0),
		 FinalAtts = [priority([Priority1])|Atts3],
		 atom_chars(RuleAtom,Rule),
		 compile(Rule,[rule([RuleAtom])|FinalAtts],RuleOid));
		(write("Error! No rule given!"), nl, fail)), !
      )
    )),


% Look at mm_rule_manager for details about this method. The ordered_rule_list attribute
% is taken into account only if the specificity criterion is 'on'. For the rest of
% the selection criteria look elsewhere. (refractoriness at the 'fire' method in
% mm_production_rule; recency at the 'calculate_conflict_set' method in 
% mm_production_rule_manager

    method((ordered_rule_list(global,[],[plog],plog,[[ConflictSetIn],ConflictSet]) :-
     (
	message_recipient(RuleClass),	
      	(get_specificity(on) => RuleClass -> 
            	(get_ordered_rule_list(OrderRules) => RuleClass,
             	 intersection(OrderRules,ConflictSetIn,ConflictSet));
	    	ConflictSet = ConflictSetIn)
    ))),

    method((duplicate_instantiations(global,[],[plog],plog,[[ConflictSetIn],ConflictSet]) :- 
	  		tidy_production_rules(ConflictSetIn,ConflictSet)
    )),

    method((scheduling(global,[],[plog],[],[Occs]) :-
     % The method calculate_conflict_set is introduced in order to
     % be able to re-define the calculation of the conflict
     % set for the production rules. See method definition above.
     (	calculate_conflict_set([Occs],RuleConflictSet) => self,
	(RuleConflictSet \== [] ->
		schedule([RuleConflictSet]) => self;
		true)
     ))),


%   +--------------------------------------------------------------------+
%   |	method CALCULATE_CONFLICT_SET
%   |	This method calculates the conflict set from the list of events
%   |   passed as an argument(Occs).
%   | This findall call build the RULE CONFLICT SET list from the list of  events
%   |  All these events have happened at the same time.
%   | The conflict set is built here in the form:
%   |		[ RuleOID-Params, RuleOID-Params, ... ]
%   | Params can have different forms as it is shown in the heading comment of this 
%   | method
%   | This has been made a separate method from Nick and is called from scheduling
%   | because it is re-defined for DEVICE in order to cater for production rules that
%   | have been already selected before and reside in the conflict_set attribute

    method((calculate_conflict_set(global,[],[plog],plog,[[Occs],ConflictSet]) :-
        (
	 message_recipient(RuleClass),
         findall( Aux,
		 (member(Occ,Occs),
		  (Occ = occurrence_tuple(Event,unique(ParamsTuple)) ->
			Unique = true;
			Occ = occurrence_tuple(Event,ParamsTuple),
			Unique = false),
		  % The event can either be primitive or complex
		  (my_instance_of(Event,mm_inheritable_ms_event) -> 
			(((((get_activated_rules(RuleOid) => Event,    % If it is primitive then it can
			     PosNeg = positive);                       % be positive or negative depending
			    (get_de_activated_rules(RuleOid) => Event, % on the (de_)activated_rules slot!
			     PosNeg = negative)),
			   transform_parameters(RuleOid,Occ,PosNeg,Occ1));  % Or it can be an update_ event!
			  (get_activated_events(AEVENTS) => Event,
			   [EventOID,AntiEventOID] = AEVENTS,
			   transform_update_parameters_rule([EventOID,AntiEventOID],RuleOid,Occ,[Ptuple,Ntuple]),
			   (Occ1 = Ntuple; Occ1 = Ptuple))),
			 Occ1 = occurrence_tuple(_,NewParamsTuple),
			 NewParamsTuple =.. [PosNeg,NewParams]);
			(ParamsTuple =.. [PosNeg,NewParams],
			 (PosNeg == positive ->			    % If it is complex it can be pos or neg
		  		get_activated_rules(RuleOid) => Event; % depending on the event occurence
				get_de_activated_rules(RuleOid) => Event))),
		  % Only instances of its own class are placed in the conflict set!
		  message_recipient(RuleManager),
		  instance_of(RuleOid,RuleManager1),
		  same_conflict_set(RuleManager1,RuleManager),
		  (get_in_token(In_token) => RuleOid -> 
			true;
			In_token = []),
                  % Check if the event occurrence is related to the rule
		  my_subset(In_token,NewParams),
		  InParamsTuple1 =.. [PosNeg,In_token],
		  (Unique = true ->
		  	InParamsTuple = unique(InParamsTuple1);
			InParamsTuple = InParamsTuple1),
		  Aux = RuleOid-InParamsTuple
		  ),
	     DifferentialConflictSet),

	 (DifferentialConflictSet \== [] -> 
			% Here the stored conflict set is accessed!
			(get_conflict_set(OldConflictSet) => RuleClass -> 
				( % Decide about the recency strategy. If recency=on then put in front of the
				  % production rule conflict set the newly added productions!
	 	 		 (get_recency(on) => RuleClass -> 
					(reverse(DifferentialConflictSet,DCS1),
					 append(DCS1,OldConflictSet,CS1),
					 duplicate_instantiations([CS1],CS2) => RuleClass,
					 reverse(CS2,ConflictSet));
					(append(OldConflictSet,DifferentialConflictSet,CS1),
					 reverse(CS1,CS2),
				 	 duplicate_instantiations([CS2],ConflictSet) => RuleClass)),
		 		 % The conflict_set is updated for DEVICE
				 (ConflictSet \== OldConflictSet -> 
		    	 	 	update_conflict_set([OldConflictSet,ConflictSet]) => RuleClass;
					true),
				 (get_spy_conflict_set(on) => RuleClass ->
					(my_subtract(ConflictSet,OldConflictSet,DiffCS),
					 store_diff_cs(DiffCS,RuleClass));
					true));
			% Do not add the new instantiations in the CS because they will be added into the EXACT CS!
				( % Decide about the recency strategy. If recency=on then put in front of the
				  % production rule conflict set the newly added productions!
	 	 		 (get_recency(on) => RuleClass -> 
					(reverse(DifferentialConflictSet,CS1),
					 duplicate_instantiations([CS1],CS2) => RuleClass,
					 reverse(CS2,ConflictSet));
					(reverse(DifferentialConflictSet,CS2),
				 	 duplicate_instantiations([CS2],ConflictSet) => RuleClass)),
				  % Also incorporate the specificity criterion complimentary with the
				  % order_rule_list method!
		 		 % The conflict_set is updated for DEVICE
		    	 	 put_conflict_set([ConflictSet]) => RuleClass,
				 (get_spy_conflict_set(on) => RuleClass ->
					store_diff_cs(ConflictSet,RuleClass);
					true)));   
			ConflictSet = [])

    ))),


    % The 'real' scheduling is done here! This method has been separated from 
    % 'scheduling' in order to be called separately from the predicate 'next_cycle'
    % that evokes the next production cycle, by re-scheduling the stored in the conflict set
    % instatiations. This is a shortcut method from mm_generic_rule
    method((schedule(global,[],[plog],[],[RuleConflictSet]) :-
     (  message_recipient(RuleManager),
	get_scheduling_event(AbstractEvent) => RuleManager,
	AbstractOccurrence = occurrence_tuple(AbstractEvent,[none,RuleManager,[RuleConflictSet]]),
	signal([[AbstractOccurrence]]) => mm_system_abstract_event
     ))),

    method((fc(global,[],[],[],[]) :-
	(message_recipient(RuleManager),
	 update_status([stopped,running]) => RuleManager,
	 put_task([start]) => RuleManager,
	 (get_spy_conflict_set(on) => RuleManager -> 
		((get_cycle_no(OldCN) => RuleManager ->
			update_cycle_no([OldCN,1]) => RuleManager;
			put_cycle_no([1]) => RuleManager),
		 nl, nl, write('Cycle no: '), write(1), nl,
		 write('New rule instantiations:'), nl,
		 (get_previous_cs(Prev) => RuleManager ->
			 (write_list_in_lines(Prev),
			  delete_previous_cs([Prev]) => RuleManager);
			 nl));
		true),
	 evaluating_now => RuleManager,
	 stop => RuleManager)
     )),

   % This predicates signals the beginning of the processing of the next production cycle.
   method((next_cycle(global,[],[],[],[]) :-
     (  message_recipient(RuleClass),
	(get_status(stopped) => RuleClass ->
		(get_conflict_set(A) => RuleClass -> 
			(delete_conflict_set([A]) => RuleClass,
			 nl, write('Stopped!'), nl, nl);
			true);
		((get_spy_conflict_set(on) => RuleClass -> 
			(get_cycle_no(CN) => RuleClass,
		 	 NextCN is CN + 1,
		 	 nl, nl, write('Cycle no: '), write(NextCN), nl,
		 	 update_cycle_no([CN,NextCN]) => RuleClass,
			 write('New rule instantiations:'), nl,
			 (get_previous_cs(Prev) => RuleClass ->
			 	(write_list_in_lines(Prev),
			  	 delete_previous_cs([Prev]) => RuleClass);
			 	nl));
			true),
		 (get_conflict_set(A) => RuleClass -> 
			(schedule([A]) => RuleClass,
		 	 evaluating_now => RuleClass);
			evaluating_now => RuleClass)))
     ))),

   method((stop(global,[],[],[],[]) :-
     (  message_recipient(RuleClass),
	(get_status(running) => RuleClass ->
		update_status([running,stopped]) => RuleClass;
		true),
	(get_task(Task) => RuleClass ->
		delete_task([Task]) => RuleClass;
		true)
     ))),

    method((spy(global,[],[],[],[]) :-
     (  message_recipient(RuleManager),
	(get_spy_conflict_set(off) => RuleManager ->
		(update_spy_conflict_set([off,on]) => RuleManager,
		 write_list(['Now spying the conflict set of',RuleManager]));
		write_list(['The conflict set of',RuleManager,'is already being spied!']))
     ))),

    method((no_spy(global,[],[],[],[]) :-
     (  message_recipient(RuleManager),
	(get_spy_conflict_set(on) => RuleManager ->
		(update_spy_conflict_set([on,off]) => RuleManager,
		 write_list(['The conflict set of',RuleManager,'is no more spied.']));
		write_list(['The conflict set of',RuleManager,'is not being spied!']))
     )))


  ]]) => meta_production_rule_manager.
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



% This method checks whether a rule manager has an autonomous conflict set
% and then it retrieves it, or it forwards the call to its super-class, until
% an autonomous Rule Class is found. The following methods also work this way 
% for the corresponding slots, which do not have an individual existence without
% an autonomous conflict set


:- replace_method([
	(get_conflict_set(global,single,[],plog,[CS]) :- 
		(get_common_cs(no) => self -> 
			CS <= conflict_set;
			(get_is_a(Super) => self,
			 get_conflict_set(CS) => Super)))
	]) => mm_production_rule_manager.

:- put_method([
	(get_ordered_rule_list(global,single,[],plog,[RL]) :- 
		(get_common_cs(no) => self -> 
			get_ordered_rule_list(RL) => super;
			(get_is_a(Super) => self,
			 get_ordered_rule_list(RL) => Super)))
	]) => mm_production_rule_manager.

:- put_method([
	(put_ordered_rule_list(global,[],[plog],[],[RL]) :- 
		(get_common_cs(no) => self -> 
			put_ordered_rule_list([RL]) => super;
			(get_is_a(Super) => self,
			 put_ordered_rule_list([RL]) => Super)))
	]) => mm_production_rule_manager.

:- put_method([
	(update_ordered_rule_list(global,[],[plog,plog],[],[ORL,NRL]) :- 
		(get_common_cs(no) => self -> 
			update_ordered_rule_list([ORL,NRL]) => super;
			(get_is_a(Super) => self,
			 update_ordered_rule_list([ORL,NRL]) => Super)))
	]) => mm_production_rule_manager.

:- put_method([
	(delete_ordered_rule_list(global,[],[plog],[],[RL]) :- 
		(get_common_cs(no) => self -> 
			delete_ordered_rule_list([RL]) => super;
			(get_is_a(Super) => self,
			 delete_ordered_rule_list([RL]) => Super)))
	]) => mm_production_rule_manager.


:- put_method([
	(get_priority_function(global,single,[],plog,[RL]) :- 
		(get_common_cs(no) => self -> 
			get_priority_function(RL) => super;
			(get_is_a(Super) => self,
			 get_priority_function(RL) => Super)))
	]) => mm_production_rule_manager.

:- replace_method([
	(get_refractoriness(global,single,[],string,[RL]) :- 
		(get_common_cs(no) => self -> 
			RL <= refractoriness;
			(get_is_a(Super) => self,
			 get_refractoriness(RL) => Super)))
	]) => mm_production_rule_manager.

:- replace_method([
	(get_recency(global,single,[],string,[RL]) :- 
		(get_common_cs(no) => self -> 
			RL <= recency;
			(get_is_a(Super) => self,
			 get_recency(RL) => Super)))
	]) => mm_production_rule_manager.

:- replace_method([
	(get_specificity(global,single,[],string,[RL]) :- 
		(get_common_cs(no) => self -> 
			RL <= specificity;
			(get_is_a(Super) => self,
			 get_specificity(RL) => Super)))
	]) => mm_production_rule_manager.

:- replace_method([
	(get_task(global,single,[],string,[RL]) :- 
		(get_common_cs(no) => self -> 
			RL <= task;
			(get_is_a(Super) => self,
			 get_task(RL) => Super)))
	]) => mm_production_rule_manager.


%*************************************************************************
%*************************************************************************
%*************************************************************************
%			ENTITIES: CLASSES
%*************************************************************************
%*************************************************************************
%*************************************************************************

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%									 %
%	class MM_PRODUCTION_RULE
%
% The behaviour of the rules instances of this class is the simplest
% (you can see it in the values of the slots, immediate, immediate etc)
%
:- new([mm_production_rule,[
    is_a([mm_generic_rule]),
    condition_coupling([user_invoked]),
    action_coupling([immediate]),
    scheduling_mode([only_one_fired]),
    cascading_mode([immediate]),
    error_recovery_mode([ignore_this_rule]),
    common_cs([no]),


%   The priority function for the specificity filter + explicit priority numbers + control rules
    priority_function([(attributes(Atts),
	                before(Before),
			member(event_number([No_of_events]),Atts),
			(member(control([yes]),Atts) ->
				Control_priority = 0;
				Control_priority = 1),
			member(priority([Priority]),Atts),
			Total_priority is 1000 * Control_priority + 100 * Priority + No_of_events,
			(get_ordered_rule_list(OrderedRules) => self -> 
			        find_before_rule(Total_priority,OrderedRules,Before);
				Before = [])
		     )]),

%   ----------------------------------------------------------------------
    recency([on]),
    specificity([on]),
    refractoriness([on]),
    status([stopped]),
    spy_conflict_set([off]),

%   ----------------------------------------------------------------------

%   ----------------------------------------------------------------------
    slot(slot_tuple(anti_event,		global,single,total,mm_generic_event)),
    slot(slot_tuple(in_token,           global,single,total,plog)),
    slot(slot_tuple(event_number,       global,single,total,integer)),
    slot(slot_tuple(rule,         global,single,total,string)),
    slot(slot_tuple(name,         global,single,optional,string)),
    slot(slot_tuple(module,     global,single,mandatory,string)),
    slot(slot_tuple(priority,     global,single,optional,integer)),
    slot(slot_tuple(control,     global,single,mandatory,string)),
%   ----------------------------------------------------------------------

    method((get_all_events(global, set, [], mm_generic_event, [[Event|Events]]) :-
	(get_event(Event) => self,
	 (get_all_activated_by(Events) => Event -> 
		true;
		Events = [])
        ))),
    method((get_all_anti_events(global, set, [], mm_generic_event, [[Event|Events]]) :-
	(get_anti_event(Event) => self,
	 (get_all_de_activated_by(Events) => Event ->
		true;
		Events = [])
        ))),

    method((display_net(global, [], [], [], []) :-
	(message_recipient(MR),
	 get_event(Event) => MR,
	 get_event_net(Events) => Event,
	 nl, write_list(['The DEVICE network of rule ',MR,' is:']),
	 print_nested_events(0,[MR,[Event,Events]])
        ))),

    method((spy_net(global, [], [], [], []) :-
	(message_recipient(RuleOID),
	 get_event(Event) => RuleOID,
	 get_event_net(Events) => Event,
	 flatten([Event|Events],L1),
	 remove_dups(L1,L2),
	 spy_events(L2)
        ))),

    method((no_spy_net(global, [], [], [], []) :-
	(message_recipient(RuleOID),
	 get_event(Event) => RuleOID,
	 get_event_net(Events) => Event,
	 flatten([Event|Events],L1),
	 remove_dups(L1,L2),
	 no_spy_events(L2)
        ))),

    method((spy_to_depth(global, [], [integer], [], [Depth]) :-
	(message_recipient(RuleOID),
	 get_event(Event) => RuleOID,
	 get_event_net(Events) => Event,
	 spy_nested_events(0,Depth,[Event,Events])        
	))),

    method((no_spy_to_depth(global, [], [integer], [], [Depth]) :-
	(message_recipient(RuleOID),
	 get_event(Event) => RuleOID,
	 get_event_net(Events) => Event,
	 no_spy_nested_events(0,Depth,[Event,Events])        
	))),

%   +--------------------------------------------------------------------+
%   |	method	EVALUATING_CONDITION					 |
%   |	
%   |	Currently no condition check for production rules, since it is always true 
%   |
     method((evaluating_condition(global,[],[plog],plog,[[ParamsTuple],ConResult]) :-
	% The refractoriness strategy has been copied here. 
	% It could equally well fit into the executing_action method.
        % There is no other reason for putting this in here!
      ( 
        message_recipient(Rule),
	instance_of(Rule,RuleClass),
	(get_refractoriness(on) => RuleClass -> 
		(get_conflict_set(OldCS) => RuleClass,
		 delete(OldCS,Rule-ParamsTuple,NewCS),
		 update_conflict_set([OldCS,NewCS]) => RuleClass);
		true),
	ParamsTuple =.. [PosNeg,Params],
	(PosNeg == positive ->
		(get_condition(Condition) => Rule,
		 (abel_path_arg(_,Condition,   occurrence_params(Params)   )  -> true ; true),
		 (abel_path_arg(_,Condition,   condition_result(ConResult) )  -> true ; true),
		 call(Condition));
		true),
	(get_spy_conflict_set(on) => RuleClass ->
		(write_list(['Executing rule instantiation: ', Rule-ParamsTuple]), nl);
		true)

      ))),

%   +--------------------------------------------------------------------+
%   |	method	EXECUTING_ACTION					 |
%   |
%   |	Takes the piece of code stored in the action slot of the rule 
%   |	object and perform two things: first establishes the information
%   |	passing from the condition to the action and second executes
%   |	the action (through executing the piece of code taken)
%   |
     method((executing_action(global,[],[plog,plog],[],[ParamsTuple,ConResult]) :-
      (	
	message_recipient(Rule),
        ParamsTuple =.. [PosNeg,In_token],
	(PosNeg == positive ->
		(get_action(Action) => Rule,
		 (abel_path_arg(_,Action,occurrence_params(In_token))    -> true ; true),
		 (abel_path_arg(_,Action,condition_result(ConResult))  -> true ; true),
		 (call(Action) -> 
			true;
			(write('Action failed for rule instantiation: '), write(Rule-ParamsTuple), nl)));  
		true),  	% Negative instantiations are just ignored for production rules!
	instance_of(Rule,RuleClass),
	next_cycle => RuleClass	% This is needed for the next cycle of rule processing!
      )))

]]) => mm_production_rule_manager,
%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



put_method([
   (delete(global,[],[],[],[]) :-
     	(message_recipient(Rule),
	 instance_of(Rule,RuleClass),
	 delete => super,
	 % The following 3 lines delete the rule from the ordered_rule_list slot
	 % of its meta-class
	 (get_ordered_rule_list(OrderedRuleList) => Rule -> 
	 	(delete(OrderedRuleList,Rule-_,OrderedRuleList1),
	 	 update_ordered_rule_list([OrderedRuleList,OrderedRuleList1]) => Rule);
		true),
	 % The following 3 lines delete the rule from the ordered_rule_list slot
	 % of its meta-class
	 (get_conflict_set(ConflictSet) => Rule -> 
	 	(delete(ConflictSet,Rule-_,ConflictSet1),
	 	 update_conflict_set([ConflictSet,ConflictSet1]) => Rule);
		true)
    ))
]) => mm_production_rule.

%									 %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

