/* Parser for DEVICE */

% This predicate accepts a string with a DEVICE rule and returns
% its parse tree.
parse(String,Tree) :-
	transform(String,List),
	(phrase(rule(Tree1,Vars),List) ->
		true
		; write_list(['CAN NOT PARSE RULE:',List]), nl, fail),
	precompile(Vars,Tree1,Tree2), !,
	optimize(Tree2,Tree), !.



/*
% This predicate accepts a string and returns a lists of atoms/tokens
% that reside in it. It is tailored to DEVICE rules, since everything
% after the 'then' keyword is not analysed.
*/

% CASE: then (termination condition)
transform([116|[104|[101|[110|RestStr]]]],[then|[RestStr]]):- !.


% CASE: space
transform([32|RestStr],RestTokens):-
	!,
	transform(RestStr,RestTokens).

% CASE: if
transform([105|[102|RestStr]],[if|RestTokens]):-
	!,
	transform(RestStr,RestTokens).

% CASE: and
transform([97|[110|[100|RestStr]]],[and|RestTokens]):-
	!,
	transform(RestStr,RestTokens).

% CASE: not
transform([110|[111|[116|RestStr]]],[not|RestTokens]):-
	!,
	transform(RestStr,RestTokens).


% CASE: { (start of prolog code)
transform([123|RestStr],[Tokenized123|[PrologCode|RestTokens]]):-
	!,
	atom_chars(Tokenized123,[40]),		% replace { with (
	get_prolog_code(RestStr,[],PrologCode,RemainStr),
	transform(RemainStr,RestTokens).

% CASE: quoted_atom  Ex. 'Mike'
transform([39|RestStr],[Token|RestTokens]):-
	!,
	get_quoted_atom(RestStr,[],Token,RemainStr),
	transform(RemainStr,RestTokens).

% CASE: \= 
transform([92|[61|RestStr]],[TokenizedX|RestTokens]):-
	!,
	atom_chars(TokenizedX,[92,61,61]),	% replace \= with \==
	transform(RestStr,RestTokens).

% CASE: =< 
transform([61|[60|RestStr]],[TokenizedX|RestTokens]):-
	!,
	atom_chars(TokenizedX,[61,60]),
	transform(RestStr,RestTokens).
	
% CASE: >= 
transform([62|[61|RestStr]],[TokenizedX|RestTokens]):-
	!,
	atom_chars(TokenizedX,[62,61]),
	transform(RestStr,RestTokens).

% CASE: any of    <  >  (  )  .  :  ;  @  ,  =,  }  \
transform([X|RestStr],[TokenizedX|RestTokens]):-
%                 <  >  (  )  .  :  ;  @  ,  =,  }  \
	member(X,[60,62,40,41,46,58,59,64,44,61,125,92]),
	!,
	(X=\=125 -> XX=X ; XX=41),	% replace }=125 with )=41
	atom_chars(TokenizedX,[XX]),
	transform(RestStr,RestTokens).


% CASE: atom
transform(Str,[Token|RestTokens]):-
	get_atom(Str,[],Token,RemainStr),
	transform(RemainStr,RestTokens).



% if (current char is . AND current atom is number) then keep on getting atom (it is a float)
get_atom([46|RestStr],CurAtom,Token,RemainStr):-
	number_chars(_,CurAtom),
	!,
	get_atom(RestStr,[46|CurAtom],Token,RemainStr).	
	
% termination condition: space < > , . : ; ( ) @ = { \
get_atom([X|RestStr],CurAtom,Token,[X|RestStr]):-
	member(X,[32,60,62,44,46,58,59,40,41,64,61,123,92]),
	!,
	reverse(CurAtom,Reversed),
	string2atom_or_var(Reversed,Token).
get_atom([X|RestStr],CurAtom,Token,RemainStr):-
	!,
	get_atom(RestStr,[X|CurAtom],Token,RemainStr).


get_quoted_atom([39|RestStr],CurQuotedAtom,Token,RestStr):-
	!,
	reverse(CurQuotedAtom,Reversed),
	atom_chars(Token,Reversed).
get_quoted_atom([X|RestStr],CurQuotedAtom,Token,RemainStr):-
	get_quoted_atom(RestStr,[X|CurQuotedAtom],Token,RemainStr).




% 125=} I push back the } char to get it as atom right after
get_prolog_code([125|RestStr],CurPrologCode,PrologCode,[125|RestStr]):-
	!,
% try as charlist
	reverse(CurPrologCode,PrologCode).

get_prolog_code([X|RestStr],CurPrologCode,PrologCode,RemainStr):-
	get_prolog_code(RestStr,[X|CurPrologCode],PrologCode,RemainStr).
	
% Token starts with an upper case letter (A-Z) ==> it is a variable
string2atom_or_var([X|Xs],var(Var)):-
	X>64, X<91,
	!,
	atom_chars(Var,[X|Xs]).
% Token starts with a digit (0-9) ==> it is a number
string2atom_or_var([X|Xs],Number):-
	X>47, X<57,
	!,
	number_chars(Number,[X|Xs]).
% any other char
string2atom_or_var(String,Token):-
	atom_chars(Token,String).


read_all_tokens(Stream,Return) :-
	read_token(Stream,String,Class),
	transform_token(String,Class,Token),
	\+ member(Token,[then,prolog]), !,
	(Token == end_of_file ->
		Return = [];
		(Return = [Token|Rest],
		 read_all_tokens(Stream,Rest))).
read_all_tokens(Stream,[prolog,'(',Goal,')'|Rest]) :- 
	read_string(Stream,"",1,"{"), !,
	read_string(Stream,"}",_,Goal),
	read_all_tokens(Stream,Rest).
read_all_tokens(Stream,[then,Action]) :- 
	read_string(Stream,"",_,Action).

transform_token(String,var,var(Var)) :- !,
	atom_chars(Var,String).
transform_token(Token,_,Token) :- 
	constant(Token), !.
transform_token(String,_,Token) :- 
	atom_chars(Token,String).


% The grammar of DEVICE rules in DCG

rule([Cond,Action],Vars) --> 
	[if], condition(Cond,CondVars,PrologVars), 
	[then], action(CondVars,PrologVars,Action),
	{union(CondVars,PrologVars,Vars)}.



	
condition(Cond,CondVars,PrologVars) --> 
	inter_object_pattern(Cond,[],CondVars,PrologVars).

inter_object_pattern(ALL_IOPS,TempCondVars,AllCondVars,PrologVars) --> 
	condition_element(CE,TempCondVars,CondVars,PrologVars1), 
	([and] ->
		inter_object_pattern(IOPs,CondVars,RestCondVars,PrologVars),
		{union(CondVars,RestCondVars,AllCondVars),
		 ALL_IOPS = [CE|IOPs]};
		{AllCondVars = CondVars,
		 PrologVars = PrologVars1,
		 ALL_IOPS = [CE]}).



condition_element(nested_neg(IOPs),TempCondVars,[],[]) --> 
	[not], ['('], !, inter_object_pattern(IOPs,TempCondVars,_,_), [')'].
condition_element(neg(IOP),TempCondVars,[],[]) --> 
	[not], !, intra_object_pattern(neg,IOP,TempCondVars,_).
condition_element(prolog(Goal),_,[],PrologVars) -->  
	[prolog], !, ['('], 
	[Goal], [')'],
	{
	   %FOTIS: string(Goal),
	   var_list(Goal,PrologVars)}.
condition_element(pos(IOP),TempCondVars,Vars,[]) --> 
	intra_object_pattern(pos,IOP,TempCondVars,Vars), !.




intra_object_pattern(Status,IOP,TempCondVars,[V|RestVars]) --> 
	var(V),
	['@'], !, class(Class),
	(['('] ->
		attribute_patterns(Status,Class,APs,[V|TempCondVars],RestVars), [')'];
		{APs = []}),
	(['\\=='] ->
		var(V1), {IOP = [V\==V1,Class,APs]};
		{IOP = [V,Class,APs]}).
intra_object_pattern(Status,[Class,Meta_Class,APs],TempCondVars,Vars) --> 
	class(Class),
	{instance_of(Class,Meta_Class)}, !,
	['('], attribute_patterns(Status,Meta_Class,APs,TempCondVars,Vars), [')'].



attribute_patterns(Status,Class,[AP|APs],TempCondVars,AllVars) --> 
	attribute_pattern(Status,Class,AP,TempCondVars,Vars), [','], !,
	{append(Vars,TempCondVars,NextTempCondVars)},
	attribute_patterns(Status,Class,APs,NextTempCondVars,RestVars),
	{union(Vars,RestVars,AllVars)}.
attribute_patterns(Status,Class,[AP],TempCondVars,Vars) --> 
	attribute_pattern(Status,Class,AP,TempCondVars,Vars).

attribute_pattern(Status,Class,defcomp(AF,var(V),Preds),TempCondVars,[V]) --> 
	attribute_function(Class,[],AF), 
	[':'], var(V), predicates(Log,Ps), !,
	{check_safety(Status,TempCondVars,V),
	 compose_pred(Log,Ps,Preds)}.
attribute_pattern(Status,Class,def(AF,var(Var)),TempCondVars,[V]) --> 
	attribute_function(Class,[],AF), [':'], 
	([unique,'('] ->
		var(V), [')'], 
		{Var = unique(V)};
		var(V),
		{Var = V}),
	!,
	{check_safety(Status,TempCondVars,V)}. 
attribute_pattern(_Status,Class,comp(AF,Preds),_TempCondVars,[]) --> 
	attribute_function(Class,[],AF),
	predicates(Log,Ps),
	{compose_pred(Log,Ps,Preds)}.



predicates(NLog,[pred(R,Val)|Ps]) --> 
	rel_operator(R),
	const(Val),
	logical_connective(Log), !,
	predicates(PLog,Ps),
	{next_log(PLog,Log,NLog)}.
predicates(_,pred(R,var(V))) --> 
	rel_operator(R), 
	var(V), !.
predicates(single,[pred(R,Val)]) --> 
	rel_operator(R), 
	const(Val).




logical_connective(and) --> 
	['&'], !.
logical_connective(or) --> 
	[';'].



rel_operator(R) --> 
	[R], {relational_operator(R), !}.



const(C) --> 
	[C], {constant(C)}.



var(V) --> 
	[var(V)], {string_var(V)}.




attribute_function(Class,TempAttrs,[Attr|Rest]) --> 
	[Attr], {atom(Attr)}, ['.'], !,
	attribute_function(Class,[Attr|TempAttrs],Rest).
attribute_function(Class,RestAttrs,[Attr]) --> 
	attribute(Class,RestAttrs,Attr).



attribute(Class,RestAttrs,Attr) --> 
	[Attr], 
	{atom(Attr), 
	 check_attribute(Attr,RestAttrs,Class)}.


class(C) --> [C], {((atom(C), get(C) => object) ->
			true;
			write_list(['No such class',C,'exists!']))}.


action(CondVars,PrologVars,Action) --> 
	[Action],
	{
	 var_list(Action,ActionVars),
	 union(CondVars,PrologVars,AllVars),
	 (my_subset(ActionVars,AllVars) ->
		true; 
		(nl, write('Error: The rule is not safe!'), nl, fail))}.









% Auxiliary information and predicates for the parser
relational_operator('=').
relational_operator('>').
relational_operator('>=').
relational_operator('=<').
relational_operator('<').
% relational_operator('\\=').
relational_operator('\\==').		% ( eclipse \\= ) <==> ( sicstus \\== )

constant(C) :- atom(C), !.
constant(C) :- integer(C), !.
constant(C) :- float(C).


string_var(Atom):-
	atom_chars(Atom,String),
	append(String,[46],TermStr), !,
	read_from_chars(TermStr,Term),
	var(Term).


next_log(single,Log,Log) :- !.
next_log(PLog,PLog,PLog).

compose_pred(Log,Ps,Ps) :- 
	var(Log), !.
compose_pred(Log,Ps,Preds) :- 
	Preds =.. [Log,Ps].

check_safety(neg,TempCondVars,V) :- 
	member(V,TempCondVars), !.
check_safety(neg,_TempCondVars,V) :- !,
	nl, 
	write_list(['The rule is not safe because var',V,
		    'is defined in a negative pattern!']).
check_safety(pos,_,_).

check_attribute(Attrib,[],Class) :- 
	path(Class,Supers),
	check_attribute_aux(Attrib,[Class|Supers],_), !.
check_attribute(Attrib,[PrevAttr|RestAttrs],Class) :- 
	path(Class,Supers),
	check_attribute_aux(Attrib,[Class|Supers],NextClass),
	check_attribute(PrevAttr,RestAttrs,NextClass).

check_attribute_aux(Attrib,[Class|_],Type) :- 
	get_slot_desc(slot_desc_tuple(Attrib,_,_,_,_,Type)) => Class, !.
check_attribute_aux(Attrib,[_|Supers],Type) :- 
	check_attribute_aux(Attrib,Supers,Type).
